import os
import logging
from uuid import UUID
from enum import Enum
from concurrent.futures import ThreadPoolExecutor

logging.basicConfig(
    level=os.getenv("LQS_LOG_LEVEL") or logging.INFO,
    format="%(asctime)s  (%(levelname)s - %(name)s): %(message)s",
)
logger = logging.getLogger(__name__)


class S3Resource(str, Enum):
    extraction = "extraction"
    ingestion = "ingestion"
    log = "log"
    record = "record"
    topic = "topic"


class Utils:
    def __init__(self, getter, s3, gen, config):
        self._getter = getter
        self._s3 = s3
        self._gen = gen
        self._config = config
        self._cached_ingestions = {}
        self._cached_message_types = {}

        if self._config.get("verbose"):
            logger.setLevel(logging.DEBUG)

    def upload_part(
        self,
        resource: S3Resource,
        resource_id: UUID,
        file_path: str,
        upload_id: str,
        part_number: int,
        key: str = None,
        part_size: int = 5 * 1024 * 1024,
    ):
        file = open(file_path, "rb")
        file.seek((part_number - 1) * part_size)
        part_data = file.read(part_size)

        logger.debug(f"Uploading part {part_number} ({len(part_data)} bytes)")

        if key is None:
            key = file_path.split("/")[-1]

        r_headers, r_params, r_body = self._s3.upload_part(
            resource=resource,
            resource_id=resource_id,
            key=key,
            part_number=part_number,
            upload_id=upload_id,
            body=part_data,
        )
        e_tag = r_headers["ETag"]
        return {"PartNumber": part_number, "ETag": e_tag}

    def upload(
        self,
        resource: S3Resource,
        resource_id: UUID,
        file_path: str,
        key: str = None,
        part_size: int = 5 * 1024 * 1024,
        max_workers: int = 8,
        exist_ok: bool = False,
    ):
        if key is None:
            key = file_path.split("/")[-1]

        if not exist_ok:
            r_headers, r_params, r_body = self._s3.head_object(
                resource=resource, resource_id=resource_id, key=key
            )
            logger.debug(f"HEAD {resource} ({resource_id}): {key} - {r_body}")
            if r_body["exists"]:
                raise FileExistsError(f"File already exists in S3: {key}")

        logger.debug("Uploading %s to %s", file_path, key)
        logger.debug("Creating multipart upload for %s %s", resource, resource_id)
        (
            _create_mpu_headers,
            _create_mpu_params,
            create_mpu_body,
        ) = self._s3.create_multipart_upload(
            resource=resource, resource_id=resource_id, key=key
        )
        try:
            upload_id = create_mpu_body["InitiateMultipartUploadResult"]["UploadId"]
            logger.debug("Upload ID: %s", upload_id)
        except KeyError as e:
            logger.error("Error creating multipart upload: %s", create_mpu_body)
            raise e

        file = open(file_path, "rb")
        file_size = os.fstat(file.fileno()).st_size
        part_count = (file_size // part_size) + 1
        logger.debug("File size: %s", file_size)
        logger.debug("Part size: %s", part_size)

        parts = []
        futures = []
        logger.debug("Uploading parts with %s workers", max_workers)
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            for i in range(part_count):
                part_number = i + 1
                future = executor.submit(
                    self.upload_part,
                    resource=resource,
                    resource_id=resource_id,
                    file_path=file_path,
                    upload_id=upload_id,
                    part_number=part_number,
                    key=key,
                    part_size=part_size,
                )
                futures.append(future)

        for future in futures:
            parts.append(future.result())
            logger.debug("Uploaded part %s of %s", len(parts), part_count)

        logger.debug("Completing multipart upload")
        (
            _complete_mpu_headers,
            _complete_mpu_params,
            complete_mpu_body,
        ) = self._s3.complete_multipart_upload(
            resource=resource,
            resource_id=resource_id,
            key=key,
            upload_id=upload_id,
            parts=parts,
        )
        logger.debug("Completed multipart upload %s", complete_mpu_body)

    def get_message_data_from_record(
        self, record, s3_bucket=None, s3_key=None, ingestion_id=None
    ):
        format = record["format"]
        if format != "ros":
            raise NotImplementedError(f"Format {format} not supported")

        if ingestion_id is None:
            ingestion_id = record["ingestion_id"]
            if ingestion_id in self._cached_ingestions:
                ingestion = self._cached_ingestions[ingestion_id]
            elif ingestion_id is not None:
                ingestion = self._getter.ingestion(ingestion_id)["data"]
                self._cached_ingestions[ingestion_id] = ingestion

        if s3_bucket is None:
            s3_bucket = record["s3_bucket"]
            if s3_bucket is None:
                s3_bucket = ingestion["s3_bucket"]
                if s3_bucket is None:
                    raise ValueError("No S3 bucket specified")

        if s3_key is None:
            s3_key = record["s3_key"]
            if s3_key is None:
                s3_key = ingestion["s3_key"]
                if s3_key is None:
                    raise ValueError("No S3 key specified")

        message_data = self._s3.get_message_data_from_record(
            record=record, s3_bucket=s3_bucket, s3_key=s3_key, ingestion_id=ingestion_id
        )

        return message_data

    def get_ros_message_from_record(
        self,
        record,
        s3_bucket=None,
        s3_key=None,
        ingestion_id=None,
        message_type_name=None,
        message_type_id=None,
    ):
        message_data = self.get_message_data_from_record(
            record=record,
            s3_bucket=s3_bucket,
            s3_key=s3_key,
            ingestion_id=ingestion_id,
        )

        # TODO: this can certainly be condensed
        if message_type_name is None:
            message_type = self._getter.message_type(message_type_id)["data"]
            message_type_name = message_type["name"]
            self._cached_message_types[message_type_id] = message_type_name
        elif message_type_id is not None:
            if message_type_id in self._cached_message_types:
                message_type_name = self._cached_message_types[message_type_id]
            else:
                message_type = self._getter.message_type(message_type_id)["data"]
                message_type_name = message_type["name"]
                self._cached_message_types[message_type_id] = message_type_name
        else:
            message_type_id = record["message_type_id"]
            if message_type_id in self._cached_message_types:
                message_type_name = self._cached_message_types[message_type_id]
            else:
                message_type = self._getter.message_type(message_type_id)["data"]
                message_type_name = message_type["name"]
                self._cached_message_types[message_type_id] = message_type_name

        message_class = self._gen.get_message_class(message_type_name)
        message = message_class()
        message.deserialize(message_data)
        return message



import os
import time
import logging
import threading

import requests
import rospy
from std_msgs.msg import String
from sensor_msgs.msg import Image, CompressedImage, CameraInfo

logging.basicConfig(
    level=os.getenv("LQS_LOG_LEVEL") or logging.INFO,
    format="%(asctime)s  (%(levelname)s - %(name)s): %(message)s",
)
logger = logging.getLogger(__name__)

# http://wiki.ros.org/ROS/Tutorials/WritingPublisherSubscriber%28python%29


class Node:
    def __init__(self, getter, lister, utils, config):
        self._get = getter
        self._list = lister
        self._utils = utils
        self._config = config

        self._log = None
        self._start_time = None
        self._end_time = None

        self._publishers = {}
        self._subscriptions = {}
        self._remaps = {}

        self._use_s3_directly = True
        # self._use_s3_directly = False

        if self._config.get("verbose"):
            logger.setLevel(logging.DEBUG)

    def get_message_class(self, message_type_name):
        if message_type_name == "sensor_msgs/Image":
            return Image
        elif message_type_name == "sensor_msgs/CompressedImage":
            return CompressedImage
        elif message_type_name == "sensor_msgs/CameraInfo":
            return CameraInfo
        else:
            logger.info(
                "Currently only supports Image, CompressedImage, and CameraInfo message types."
            )
            raise Exception(f"Unsupported message type: {message_type_name}")

    def subscribe(self, log_id, topic_names, rel_start_time, rel_end_time, remaps={}):
        rospy.init_node("lqs_client", anonymous=True)

        self._log = self._get.log(log_id=log_id)["data"]

        self._start_time = self._log["start_time"] + rel_start_time
        self._end_time = self._log["start_time"] + rel_end_time

        topics_all = self._list.topics(log_id=log_id, limit=100)["data"]
        self._remaps = remaps
        self._subscriptions = {}
        for topic in topics_all:
            if topic["name"] in topic_names:
                message_class = self.get_message_class(topic["message_type_name"])
                self._publishers[topic["name"]] = rospy.Publisher(
                    remaps.get(topic["name"], topic["name"]),
                    message_class,
                    queue_size=10,
                )
                thread = threading.Thread(
                    target=self.load_records, args=(topic["name"],)
                )
                thread.start()
                self._subscriptions[topic["name"]] = {
                    "topic": topic,
                    "thread": thread,
                    "has_data": True,  # assume there is data to start
                    "next_message_timestamp": None,
                    "records": [],
                }

    def fetch_record_bytes(self, record):
        start_time = time.time()
        if self._use_s3_directly:
            mode = "direct"
            record["bytes"] = self._utils.get_message_data_from_record(record)
        else:
            mode = "presgined"
            record["bytes"] = requests.get(record["bytes_url"]).content
        end_time = time.time()
        print(
            f"({mode}) fetched bytes for {record['timestamp']} in {end_time - start_time}"
        )

    def load_records_old(self, topic_name):
        # limit = 100 if self._use_s3_directly else 10
        limit = 10
        offset = 0
        while True:
            if topic_name not in self._subscriptions:
                continue
            sub = self._subscriptions[topic_name]
            records = sub["records"]
            if len(records) >= limit:
                continue
            topic = sub["topic"]
            records_response = self._list.records(
                log_id=self._log["id"],
                topic_id=topic["id"],
                timestamp_gte=self._start_time,
                timestamp_lt=self._end_time,
                limit=limit,
                offset=offset,
                include_bytes=False if self._use_s3_directly else True,
            )
            records = records_response["data"]
            print(f"loaded {len(records)} records for {topic_name}")
            if len(records) == 0:
                self._subscriptions[topic_name]["has_data"] = False
                break
            offset += limit
            # fetch message data
            for record in records:
                # print(f"fetching bytes for {topic_name} ({record['timestamp']})")
                record[
                    "bytes"
                ] = None  # init to none to avoid race condition in publish check
                # threading.Thread(target=self.fetch_record_bytes, args=(record,)).start()
                self.fetch_record_bytes(record)
            # combine records, sort by timestamp
            self._subscriptions[topic_name]["records"] += records
            self._subscriptions[topic_name]["records"].sort(
                key=lambda x: x["timestamp"]
            )
            next_record = self._subscriptions[topic_name]["records"][0]
            self._subscriptions[topic_name]["next_message_timestamp"] = next_record[
                "timestamp"
            ]

    def load_records(self, topic_name):
        # limit = 100 if self._use_s3_directly else 10
        limit = 100
        offset = 0
        while True:
            if topic_name not in self._subscriptions:
                continue
            sub = self._subscriptions[topic_name]
            records = sub["records"]
            # if len(records) > limit:
            #     continue
            topic = sub["topic"]
            records_response = self._list.records(
                log_id=self._log["id"],
                topic_id=topic["id"],
                timestamp_gte=self._start_time,
                timestamp_lt=self._end_time,
                limit=limit,
                offset=offset,
                include_bytes=False if self._use_s3_directly else True,
            )
            records = records_response["data"]
            print(f"loaded {len(records)} records for {topic_name}")
            if len(records) == 0:
                # self._subscriptions[topic_name]["has_data"] = False
                break
            offset += limit
            # fetch message data
            # for record in records:
            #     record["bytes"] = self._utils.get_message_data_from_record(record)
            #     self._subscriptions[topic_name]["records"].append(record)
            self._subscriptions[topic_name]["records"] += records
            next_record = self._subscriptions[topic_name]["records"][0]
            self._subscriptions[topic_name]["next_message_timestamp"] = next_record[
                "timestamp"
            ]

    def publish_next_message(self):
        # look through our records and publish the next message based on timestamp
        next_message_time = None
        next_message_topic = None
        for topic_name in self._subscriptions:
            sub = self._subscriptions[topic_name]
            if sub["next_message_timestamp"] is None:
                if sub["has_data"]:
                    # we haven't loaded data for this topic yet, so we can't publish anything
                    return
                else:
                    # we've loaded all the data for this topic, so we can skip it
                    continue
            if (
                next_message_time is None
                or sub["next_message_timestamp"] < next_message_time
            ):
                # this topic may have the next message to publish
                next_message_time = sub["next_message_timestamp"]
                next_message_topic = topic_name

        if next_message_topic is not None:
            self.publish(next_message_topic)
            return True
        else:
            return False

    def publish(self, topic_name):
        sub = self._subscriptions[topic_name]
        topic = sub["topic"]
        records = sub["records"]
        if len(records) == 0:
            self._subscriptions[topic_name]["has_data"] = False
            return
        record = records.pop(0)
        # TODO: prefetch bytes for next record

        # message_data = requests.get(record["bytes_url"]).content
        # while record["bytes"] is None:
        #     time.sleep(0.01)
        # message_data = record["bytes"]
        message_data = self._utils.get_message_data_from_record(record)
        message_class = self.get_message_class(topic["message_type_name"])
        message = message_class()
        message.deserialize(message_data)
        publisher = self._publishers[topic_name]
        # print(f"publishing {topic_name} ({record['timestamp']}) {len(message_data)} bytes")
        publisher.publish(message)
        if len(self._subscriptions[topic_name]["records"]) == 0:
            print(f"no more records for {topic_name}")
            self._subscriptions[topic_name]["has_data"] = False
            return
        next_record = self._subscriptions[topic_name]["records"][0]
        self._subscriptions[topic_name]["next_message_timestamp"] = next_record[
            "timestamp"
        ]

    def play_old(self, rate=10):
        rate = rospy.Rate(rate)
        while not rospy.is_shutdown():
            self.publish_next_message()
            data_remaining = False
            for topic_name in self._subscriptions:
                if self._subscriptions[topic_name]["has_data"]:
                    data_remaining = True
                    break
            if not data_remaining:
                logger.info("No more data to publish.")
                print("No more data to publish.")
                break
            rate.sleep()

    def play(self, rate=10):
        rate = rospy.Rate(rate)
        while not rospy.is_shutdown():
            self.publish_next_message()
            data_remaining = False
            for topic_name in self._subscriptions:
                if self._subscriptions[topic_name]["has_data"]:
                    data_remaining = True
                    break
            if not data_remaining:
                logger.info("No more data to publish.")
                print("No more data to publish.")
                break
            rate.sleep()
